(*
 *  This unit is copyright Matt Harrison 1998-2005.
 *  The code is free for non-commercial use.  If commercial use is required then
 *  please contact matt@lummie.co.uk for details of use.  There maybe a small cost depending on the use of the unit
 *  This Software is distributed on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, either expressed or implied.
 *
 *  Description
 *  -----------
 *  TTypedFiler is similar to TFiler as in it can be assigned to a Stream for Type Checked reading and writing.
 *  It has several benefits over the TFiler based objects.
 *      1) All information is stored in a binary form
 *      2) it can write a variant correctly - the delphi source has problems with reading and writing smallint/shortint variant types
 *      3) it can read/ write multi-dimensional variant arrays and custom variant types
 *      4) it supports reading and writing compressed streams
 *      5) it supports reading and writing encrypted streams
 *      6) it supports reading and writing pictures
 *      7) it is TInterfacedObject based so simply assign the create to a ITypedFiler interface, use it and it will free itself
 *
 *
 *  History
 *  -------
 *  Version 1.0.0 - release to public.
 *)




unit TypedFiler;

interface

//{$DEFINE TurboPowerLockBox} // tpabbrevia.sourceforge.net - Encryption Support
//{$DEFINE TurboPowerAbbrevia} // tplockbox.sourceforge.net - ZIP Comrpession support
//{$DEFINE PNGImage}    // pngdelphi.sourceforge.net
//{$DEFINE RegisterGraphic} // if PNGImage is enabled ensures that the PNG format is registered with TPicture.

uses
  Classes, SysUtils,  Graphics;

type
  TS_CONST_TYPE = byte;

type
  TIPAddress = record
    I1: Byte;
    I2: Byte;
    I3: Byte;
    I4: Byte;
  end;


const
  EMPTY_GUID : TGUID = '{00000000-0000-0000-0000-000000000000}';

const
  TS_EMPTY : TS_CONST_TYPE = $00;
  TS_CARDINAL : TS_CONST_TYPE = $01;
  TS_SMALLINT : TS_CONST_TYPE = $02;
  TS_STRING : TS_CONST_TYPE = $03;
  TS_COLOR : TS_CONST_TYPE = $04;
  TS_GUID : TS_CONST_TYPE = $05;
  TS_WIDESTRING : TS_CONST_TYPE = $06;
  TS_BOOLEAN : TS_CONST_TYPE = $07;
  TS_FLOAT : TS_CONST_TYPE = $08;
  TS_SINGLE : TS_CONST_TYPE = $09;
  TS_CURRENCY : TS_CONST_TYPE = $0A;
  TS_DATETIME : TS_CONST_TYPE = $0B;
  TS_VARIANT : TS_CONST_TYPE = $0C;
  TS_WORD : TS_CONST_TYPE = $0D;
  TS_INTEGER : TS_CONST_TYPE = $0E;
  TS_EXCEPTION : TS_CONST_TYPE = $0F;
  TS_STREAM : TS_CONST_TYPE = $10;
  TS_INT64 : TS_CONST_TYPE = $11;
  TS_IPADDRESS : TS_CONST_TYPE = $12;
  TS_VERSION : TS_CONST_TYPE = $13;
  TS_ENCRYPTEDSTREAM : TS_CONST_TYPE = $14;
  TS_VARTYPE : TS_CONST_TYPE = $15;
  TS_SHORTINT : TS_CONST_TYPE = $16;
  TS_BYTE : TS_CONST_TYPE = $17;
  TS_FONT : TS_CONST_TYPE = $18;
  TS_PICTURE : TS_CONST_TYPE = $19;
  TS_COMPRESSEDSTREAM : TS_CONST_TYPE  = $1A;

type
  TDynamicByteArray = array of byte;

  ITypedFiler = interface (IInterface)
    ['{5E1ACCF9-B894-4969-8CAF-7A8B4DD9DE77}']
    procedure Clear;
    procedure GotoEnd;
    procedure GotoStart;
    function ReadBoolean: Boolean;
    function ReadByte: Byte;
    function ReadCardinal: Cardinal;
    function ReadCurrency: Currency;
    function ReadColor: TColor;
    function ReadDateTime: TDateTime;
    procedure ReadEncryptedStream(Into : TStream; const Key : string);
    procedure ReadCompressedStream(Into : TStream);
    function ReadException: Exception;
    function ReadFloat: Extended;
    procedure ReadFont(AFont : TFont);
    function ReadGUID: TGUID;
    function ReadInt64: Int64;
    function ReadInteger: Integer;
    function ReadIPAddress: TIPAddress;
    function ReadNextType: TS_CONST_TYPE;
    function ReadPicture: TPicture;
    function ReadShortInt: ShortInt;
    function ReadSingle: Single;
    function ReadSmallInt: SmallInt;
    procedure ReadStream(Into : TStream);
    function ReadString: string;
    function ReadVariant: Variant;
    function ReadVarType: TVarType;
    function ReadVersion: Cardinal;
    function ReadWideString: widestring;
    function ReadWord: Word;

    function RawRead(NumberOfBytes : integer) : TDynamicByteArray;
    procedure RawWrite(Data : TDynamicByteArray);

    procedure WriteBoolean(const ABoolean : Boolean);
    procedure WriteByte(const AValue : Byte);
    procedure WriteCardinal(const AValue : Cardinal);
    procedure WriteCurrency(const AValue: Currency);
    procedure WriteColor(const AValue: TColor);
    procedure WriteDateTime(const AValue: TDateTime);
    procedure WriteEncryptedStream(AStream : TStream; FromStart : boolean; const Key : string);
    procedure WriteCompressedStream(AStream : TStream; FromStart : boolean);
    procedure WriteException(msg : string); overload;
    procedure WriteException(e : Exception); overload;
    procedure WriteFloat(const AValue: Extended);
    procedure WriteFont(AFont : TFont);
    procedure WriteGUID(const AGUID : TGUID);
    procedure WriteInt64(AInt64 : Int64);
    procedure WriteInteger(const AValue : integer);
    procedure WriteIPAddress(IPAddress : TIPAddress);
    procedure WritePicture(APicture : TPicture);
    procedure WriteShortInt(const AValue : ShortInt);
    procedure WriteSingle(const AValue: Single);
    procedure WriteSmallInt(const AValue : SmallInt);
    procedure WriteStream(AStream : TStream; FromStart : boolean);
    procedure WriteString(const AString : string);
    procedure WriteVariant(const AValue: Variant);
    procedure WriteVarType(AVarType : TVarType);
    procedure WriteVersion(AValue : Cardinal);
    procedure WriteWideString(const AWideString : WideString);
    procedure WriteWord(const AValue : Word);
  end;



  TTypedFiler = class(TInterfacedObject, ITypedFiler)
  private
    FStream : TStream;
  public
    constructor Create(AStream : TStream); reintroduce;
    destructor Destroy; override;
    procedure Clear;
    procedure GotoEnd; virtual;
    procedure GotoStart; virtual;
    function TypedFilerTypeToString(AType : TS_CONST_TYPE) : string;
    function ReadBoolean: Boolean; virtual;
    function ReadByte: Byte; virtual;
    function ReadCardinal: Cardinal; virtual;
    function ReadCurrency: Currency; virtual;
    function ReadColor: TColor; virtual;
    function ReadDateTime: TDateTime; virtual;
    procedure ReadCompressedStream(Into : TStream); virtual;
    procedure ReadEncryptedStream(Into : TStream; const Key : string); virtual;
    function ReadException: Exception; virtual;
    function ReadFloat: Extended; virtual;
    procedure ReadFont(AFont : TFont); virtual;
    function ReadGUID: TGUID; virtual;
    function ReadInt64: Int64; virtual;
    function ReadInteger: Integer; virtual;
    function ReadIPAddress: TIPAddress; virtual;
    function ReadNextType: TS_CONST_TYPE; virtual;
    function ReadPicture: TPicture; virtual;
    function ReadShortInt: ShortInt; virtual;
    function ReadSingle: Single; virtual;
    function ReadSmallInt: SmallInt; virtual;
    procedure ReadStream(Into : TStream); virtual;
    function ReadString: string; virtual;
    function ReadVariant: Variant; virtual;
    function ReadVarType: TVarType; virtual;
    function ReadVersion: Cardinal; virtual;
    function ReadWideString: widestring; virtual;
    function ReadWord: Word; virtual;
    procedure WriteBoolean(const ABoolean : Boolean); virtual;
    procedure WriteByte(const AValue : Byte); virtual;
    procedure WriteCardinal(const AValue : Cardinal); virtual;
    procedure WriteCurrency(const AValue: Currency); virtual;
    procedure WriteColor(const AValue: TColor); virtual;
    procedure WriteDateTime(const AValue: TDateTime); virtual;
    procedure WriteCompressedStream(AStream : TStream; FromStart : boolean); virtual;
    procedure WriteEncryptedStream(AStream : TStream; FromStart : boolean; const Key : string); virtual;
    procedure WriteException(msg : string); overload; virtual;
    procedure WriteException(e : Exception); overload; virtual;
    procedure WriteFloat(const AValue: Extended); virtual;
    procedure WriteFont(AFont : TFont); virtual;
    procedure WriteGUID(const AGUID : TGUID); virtual;
    procedure WriteInt64(AInt64 : Int64); virtual;
    procedure WriteInteger(const AValue : integer); virtual;
    procedure WriteIPAddress(IPAddress : TIPAddress); virtual;
    procedure WritePicture(APicture : TPicture); virtual;
    procedure WriteShortInt(const AValue : ShortInt); virtual;
    procedure WriteSingle(const AValue: Single); virtual;
    procedure WriteSmallInt(const AValue : SmallInt); virtual;
    procedure WriteStream(AStream : TStream; FromStart : boolean); virtual;
    procedure WriteString(const AString : string); virtual;
    procedure WriteVariant(const AValue: Variant); virtual;
    procedure WriteVarType(AVarType : TVarType); virtual;
    procedure WriteVersion(AValue : Cardinal); virtual;
    procedure WriteWideString(const AWideString : WideString); virtual;
    procedure WriteWord(const AValue : Word); virtual;
    function RawRead(NumberOfBytes : integer) : TDynamicByteArray;
    procedure RawWrite(Data : TDynamicByteArray);
  end;


implementation

uses
  Variants,
  {$IFDEF PNGImage}
  PNGImage,
  {$ENDIF}
  {$IFDEF TurboPowerLockBox}
  lbCipher, lbProc,
  {$ENDIF}
  {$IFDEF TurboPowerAbbrevia}
  AbUnzPrc, AbZipPrc,
  {$ENDIF}
  Jpeg
  ;

function TTypedFiler.TypedFilerTypeToString(AType : TS_CONST_TYPE) : string;
begin
  if (AType = TS_EMPTY) then result := 'Empty'
  else if (AType = TS_CARDINAL) then result := 'Cardinal'
  else if (AType = TS_SMALLINT ) then result := 'SmallInt'
  else if (AType = TS_STRING) then result := 'String'
  else if (AType = TS_COLOR) then result := 'Color'
  else if (AType = TS_GUID) then result := 'GUID'
  else if (AType = TS_WIDESTRING) then result := 'Widestring'
  else if (AType = TS_BOOLEAN) then result := 'Boolean'
  else if (AType = TS_FLOAT) then result := 'Float'
  else if (AType = TS_SINGLE) then result := 'Single'
  else if (AType = TS_CURRENCY) then result := 'Currency'
  else if (AType = TS_DATETIME) then result := 'DateTime'
  else if (AType = TS_VARIANT) then result := 'Variant'
  else if (AType = TS_WORD) then result := 'Word'
  else if (AType = TS_INTEGER) then result := 'Integer'
  else if (AType = TS_EXCEPTION) then result := 'Exception'
  else if (AType = TS_STREAM) then result := 'Stream'
  else if (AType = TS_INT64) then result := 'Int64'
  else if (AType = TS_IPADDRESS) then result := 'IP Address'
  else if (AType = TS_VERSION) then result := 'Version'
  else if (AType = TS_ENCRYPTEDSTREAM) then result := 'Encrypted Stream'
  else if (AType = TS_VARTYPE) then result := 'Variant Type'
  else if (AType = TS_SHORTINT) then result := 'ShortInt'
  else if (AType = TS_BYTE) then result := 'Byte'
  else if (AType = TS_FONT) then result := 'Font'
  else if (AType = TS_PICTURE) then result := 'Picture'
  else if (AType = TS_COMPRESSEDSTREAM) then result := 'Compressed Stream'
  else
    result := format('Unknow Type : $%x',[AType]);
end;


procedure TTypedFiler.Clear;
begin
  FStream.Size := 0;
end;

procedure TTypedFiler.GotoEnd;
begin
  FStream.seek(0,soFromEnd);
end;

procedure TTypedFiler.GotoStart;
begin
  FStream.seek(0,soFromBeginning);
end;

function TTypedFiler.ReadBoolean: Boolean;
var
  b: Byte;
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_BOOLEAN then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_BOOLEAN)]);
  
  FStream.ReadBuffer(b, SizeOf(b));
  result := boolean(b);
end;

function TTypedFiler.ReadByte: Byte;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_BYTE then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_BYTE)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadCardinal: Cardinal;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_CARDINAL then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_CARDINAL)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadCurrency: Currency;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_CURRENCY then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_CURRENCY)]);

  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadDateTime: TDateTime;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_DATETIME then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_DATETIME)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

{$IFDEF TurboPowerLockBox}
procedure TTypedFiler.ReadEncryptedStream(Into : TStream; const Key :
        string);
var
  ASize: Int64;
  EStream: TMemoryStream;
  ID: TS_CONST_TYPE;
  Key128: TKey128;
  IntoPos: Int64;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_ENCRYPTEDSTREAM then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_ENCRYPTEDSTREAM)]);

  ASize := ReadInt64;

  IntoPos := Into.Position;

  EStream := TMemoryStream.create;
  try
    EStream.CopyFrom(FStream,ASize);
    EStream.seek(0,soFromBeginning);
    GenerateLMDKey(Key128, SizeOf(Key128), Key);
    RDLEncryptStream(EStream,Into,Key128,Sizeof(Key128),false);
  finally
    EStream.free;
  end;
  Into.Seek(IntoPos,soFromBeginning);
end;
{$ELSE}
procedure TTypedFiler.ReadEncryptedStream(Into : TStream; const Key :
        string);
begin
  raise exception.create('TTypedFiler has not been compiled with encryption support');
end;
{$ENDIF}

function TTypedFiler.ReadException: Exception;
var
  ExceptionMessage: string;
  ID: TS_CONST_TYPE;
begin
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_EXCEPTION then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_EXCEPTION)]);

  ExceptionMessage := ReadString;
  result := exception.Create(ExceptionMessage);
end;

function TTypedFiler.ReadFloat: Extended;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_FLOAT then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_FLOAT)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

procedure TTypedFiler.ReadFont(AFont : TFont);
var
  ID: TS_CONST_TYPE;
  OldOnChange: TNotifyEvent;
  HoldIt: string;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_FONT then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_FONT)]);
  
  // store and detach on change event
  OldOnChange := AFont.OnChange;
  AFont.OnChange := nil;
  
  // read font information
  HoldIt := ReadString;
  AFont.Size := ReadInteger;
  AFont.Color := ReadInteger;
  
  AFont.style := [];
  if ReadBoolean then AFont.style := AFont.style + [fsbold];
  if ReadBoolean then AFont.style := AFont.style + [fsItalic];
  if ReadBoolean then AFont.style := AFont.style + [fsUnderline];
  if ReadBoolean then AFont.style := AFont.style + [fsStrikeOut];
  
  // reassign on change event then set name to cause on change
  AFont.OnChange := OldOnChange;
  AFont.Name := HoldIt;
end;

function TTypedFiler.ReadGUID: TGUID;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_GUID then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_GUID)]);

  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadInt64: Int64;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_INT64 then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_INT64)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadInteger: Integer;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_INTEGER then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_INTEGER)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadIPAddress: TIPAddress;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_IPADDRESS then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_IPADDRESS)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadNextType: TS_CONST_TYPE;
var
  CurrentPos: Int64;
begin
  if FStream.Position = FStream.Size then // if we are at then end of the stream then return TS_EMPTY
  begin
    result := TS_EMPTY;
  end
  else
  begin // otherwise read the next type
    CurrentPos := FStream.Position;
    try
      FStream.ReadBuffer(result,Sizeof(TS_CONST_TYPE));
    finally
      FStream.Position := CurrentPos;
    end;
  end;
end;

function TTypedFiler.ReadPicture: TPicture;
var
  GraphicStream: TMemoryStream;
  ID: TS_CONST_TYPE;
  GraphicClassName: string;
  Graphic: TGraphic;
  GraphicClass: TClass;
  NextType: TS_CONST_TYPE;
begin
  result := nil;
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');

  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_PICTURE then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_PICTURE)]);

  GraphicClassName := ReadString;

  if GraphicClassName <> '<EMPTY>' then
  begin
    GraphicClass := GetClass(GraphicClassName);
    if not assigned(GraphicClass) then raise exception.createfmt('Class not found in TypedStream.ReadPicture : %s',[GraphicClassName]);

    Graphic := TGraphicClass(GraphicClass).Create;
    if assigned(graphic) then
    begin
      try
        NextType := ReadNextType;
        if NextType = TS_STREAM then
        begin
          GraphicStream := TMemoryStream.Create;
          try
            ReadStream(GraphicStream);
            GraphicStream.Seek(0,soFromBeginning);
            Graphic.LoadFromStream(GraphicStream);
          finally
            GraphicStream.Free;
          end;
          result := TPicture.create;
          result.Assign(Graphic);
        end
        else if NextType = TS_STRING then
        begin
          readString;
          result := nil;
        end
        else
          raise exception.CreateFmt('Type %x found in Picture Stream expecting string of stream.',[NextType]);

      finally
        Graphic.Free;
      end;
    end
    else
      raise exception.createfmt('Could not create an instance of ''%s''',[GraphicClassName]);
  end;
end;

function TTypedFiler.ReadShortInt: ShortInt;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_SHORTINT then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_SHORTINT)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadSingle: Single;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_SINGLE then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_SINGLE)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadSmallInt: SmallInt;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_SMALLINT then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_SMALLINT)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

procedure TTypedFiler.ReadStream(Into : TStream);
var
  ASize: Int64;
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_STREAM then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_STREAM)]);
  
  ASize := ReadInt64;
  
  if ASize > 0 then
  begin
    Into.CopyFrom(FStream,ASize);
    Into.seek(0,soFromBeginning);
  end;
end;

function TTypedFiler.ReadString: string;
var
  ALen: Cardinal;
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_STRING then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_STRING)]);

  ALen := ReadCardinal;
  SetLength(Result, ALen);
  if ALen > 0 then
  begin
    FStream.ReadBuffer(Result[1], Length(Result));
  end;
end;


function TTypedFiler.ReadVariant: Variant;
var
  ArrayDims: Integer;
  ArrayHighBounds: array of integer;
  ArrayLowBounds: array of integer;
  ItteratorArray: array of integer;
  
      procedure ReadArray(DimNo : integer);
      var
        CellValue : variant;
        i : integer;
      begin
        if (DimNo < ArrayDims) then
        begin
          for i := ArrayLowBounds[DimNo] to ArrayHighBounds[DimNo] do
          begin
            ItteratorArray[DimNo] := i;
            if DimNo = (ArrayDims - 1) then
            begin
              Cellvalue := ReadVariant;
              VarArrayPut(result,CellValue, ItteratorArray)
            end
            else
            begin
              ReadArray(DimNo+1);
            end;
          end;
        end;
      end;
  
      function ReadCustomVariant : Variant;
      var
        CustomStream: TMemoryStream;
        CustomType: TCustomVariantType;
        CustomTypeClassName: string;
        VarStreamer: IVarStreamable;
      begin
        CustomTypeClassName := ReadString;
        CustomStream := TMemoryStream.create;
        try
          ReadStream(CustomStream);
          if not FindCustomVariantType(CustomTypeClassName, CustomType) or not Supports(TObject(CustomType), IVarStreamable, VarStreamer) then
          begin
            raise exception.create('Custom Variant of type "%s" not registered or Is not streamable in TTypedStream.ReadVariant');
          end;
          TVarData(Result).VType := CustomType.VarType;
          VarStreamer.StreamIn(TVarData(Result), CustomStream);
        finally
          FreeAndNil(CustomStream);
        end;
      end;

  var
    ID : TS_CONST_TYPE;
    VType : TVarType;

    i : integer;
    CreateArrayBounds : array of integer;

begin
  result := varEmpty;
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_VARIANT then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_VARIANT)]);

  VType := ReadVarType;
  if ((Vtype and varArray) <> 0) or ((Vtype and varByRef) <> 0) then
  begin
    //variant is an array
    ArrayDims := ReadInteger;

    // build up Low and High Bounds
    SetLength(ArrayLowBounds,ArrayDims);
    SetLength(ArrayhighBounds,ArrayDims);
    for i := 0 to ArrayDims -1 do
    begin
      ArrayLowBounds[i] := ReadInteger;
      ArrayHighBounds[i] := ReadInteger;
    end;

    // fill the CreateArrayBounds with the upper and Lower bounds
    SetLength(CreateArrayBounds,ArrayDims*2);
    for i := 0 to ArrayDims -1 do
    begin
      CreateArrayBounds[(i*2)] := ArrayLowBounds[i];
      CreateArrayBounds[(i*2)+1] := ArrayHighBounds[i];
    end;

    result := VarArrayCreate(CreateArrayBounds,(VType and varTypeMask));

    // setup itterator array
    SetLength(ItteratorArray,ArrayDims);
    // initialize to Low Bounds
    for i := 0 to ArrayDims -1 do
    begin
      ItteratorArray[i] := ArrayLowBounds[i];
    end;

    ReadArray(0);

  end
  else
  begin
    // variant is not an array
    case VType and varTypeMask of // only write basic types
      varEmpty :
      begin
        result := varEmpty;
      end;

      varNull:
      begin
        result := varNull;
      end;

      varOleStr:
      begin
        result := ReadWideString;
      end;

      varString:
      begin
        result := ReadString;
      end;

      varByte :
      begin
        result := ReadByte;
      end;

      varShortInt:
      begin
        result := ReadShortInt;
      end;

      varWord:
      begin
        result := ReadWord;
      end;

      varSmallInt:
      begin
        result := ReadSmallInt;
      end;

      varInteger:
      begin
        result := readInteger;
      end;


      varSingle:
      begin
        result := ReadSingle;
      end;

      varDouble:
      begin
        result := ReadFloat;
      end;

      varCurrency:
      begin
        result := ReadCurrency;
      end;

      varDate:
      begin
        result := ReadDateTime;
      end;

      varBoolean:
      begin
        result := ReadBoolean
      end;

      varLongWord:
      begin
        result := ReadInt64;
      end;

      varInt64:
      begin
        result := ReadInt64;
      end;
    else
      begin
        result := ReadCustomVariant;
      end;
    end;
  end;
end;

function TTypedFiler.ReadVarType: TVarType;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_VARTYPE then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_VARTYPE)]);
  FStream.ReadBuffer(result,Sizeof(result));
end;

function TTypedFiler.ReadVersion: Cardinal;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_VERSION then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_VERSION)]);
  
  FStream.ReadBuffer(Result, SizeOf(Result));
end;

function TTypedFiler.ReadWideString: widestring;
var
  ALen: Cardinal;
  ws: Widestring;
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_WIDESTRING then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_WIDESTRING)]);

  ALen := ReadCardinal;
  SetLength(ws, ALen div 2);
  if ALen > 0 then
  begin
    FStream.ReadBuffer(pointer(ws)^, ALen);
  end;
  result := ws;
end;

function TTypedFiler.ReadWord: Word;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_WORD then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_WORD)]);

  FStream.ReadBuffer(Result, SizeOf(Result));
end;


procedure TTypedFiler.WriteBoolean(const ABoolean : Boolean);
var
  b: Byte;
begin
  FStream.WriteBuffer(TS_BOOLEAN,Sizeof(TS_CONST_TYPE));
  b := byte(ABoolean);
  FStream.WriteBuffer(b, SizeOf(b));
end;

procedure TTypedFiler.WriteByte(const AValue : Byte);
begin
  FStream.WriteBuffer(TS_BYTE,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteCardinal(const AValue : Cardinal);
begin
  FStream.WriteBuffer(TS_CARDINAL,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteColor(const AValue: TColor);
begin
  FStream.WriteBuffer(TS_COLOR,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteCurrency(const AValue: Currency);
begin
  FStream.WriteBuffer(TS_CURRENCY,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteDateTime(const AValue: TDateTime);
begin
  FStream.WriteBuffer(TS_DATETIME,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

{$IFDEF TurboPowerLockBox}
procedure TTypedFiler.WriteEncryptedStream(AStream : TStream;
        FromStart : boolean; const Key : string);
var
  ASize: Int64;
  EStream: TMemoryStream;
  Key128: TKey128;
begin
  if FromStart then
  begin
    FStream.seek(0,soFromBeginning);
  end;
  
  EStream := TMemoryStream.create;
  try
    GenerateLMDKey(Key128, SizeOf(Key128), Key);
    RDLEncryptStream(AStream,EStream,Key128,Sizeof(Key128),true);
    EStream.seek(0,soFromBeginning);
    ASize := EStream.size;
    FStream.WriteBuffer(TS_ENCRYPTEDSTREAM,Sizeof(TS_CONST_TYPE));
    WriteInt64(ASize);
    FStream.CopyFrom(EStream,ASize);
  finally
    EStream.free;
  end;
end;
{$ELSE}
procedure TTypedFiler.WriteEncryptedStream(AStream : TStream;
        FromStart : boolean; const Key : string);
begin
  raise exception.create('TTypedFiler has not been compiled with encryption support');
end;
{$ENDIF}


procedure TTypedFiler.WriteException(msg : string);
var
  e: exception;
begin
  e := exception.create(msg);
  try
    WriteException(e);
  finally
    e.free;
  end;
end;

procedure TTypedFiler.WriteException(e : Exception);
begin
  FStream.WriteBuffer(TS_EXCEPTION,Sizeof(TS_CONST_TYPE));
  WriteString(e.Message);
end;

procedure TTypedFiler.WriteFloat(const AValue: Extended);
begin
  FStream.WriteBuffer(TS_FLOAT,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteFont(AFont : TFont);
begin
  FStream.WriteBuffer(TS_FONT,Sizeof(TS_CONST_TYPE));
  WriteString(AFont.name);
  WriteInteger(AFont.Size);
  WriteInteger(AFont.Color);
  WriteBoolean(fsBold in AFont.Style);
  WriteBoolean(fsItalic in AFont.Style);
  WriteBoolean(fsUnderline in AFont.Style);
  WriteBoolean(fsStrikeOut in AFont.Style);
end;

procedure TTypedFiler.WriteGUID(const AGUID : TGUID);
begin
  FStream.WriteBuffer(TS_GUID,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AGUID, SizeOf(AGUID));
end;

procedure TTypedFiler.WriteInt64(AInt64 : Int64);
begin
  FStream.WriteBuffer(TS_INT64,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AInt64, SizeOf(AInt64));
end;

procedure TTypedFiler.WriteInteger(const AValue : integer);
begin
  FStream.WriteBuffer(TS_INTEGER,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteIPAddress(IPAddress : TIPAddress);
begin
  FStream.WriteBuffer(TS_IPADDRESS, Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(IPAddress, Sizeof(IPAddress));
end;

procedure TTypedFiler.WritePicture(APicture : TPicture);
var
  GraphicStream: TMemoryStream;
begin
  FStream.WriteBuffer(TS_PICTURE,Sizeof(TS_CONST_TYPE));
  if assigned(APicture.Graphic) then
  begin
    WriteString(APicture.Graphic.ClassName);

    if not APicture.Graphic.Empty then
    begin
      GraphicStream := TMemoryStream.Create;
      try
        APicture.graphic.SaveToStream(GraphicStream);
        WriteStream(GraphicStream,true);
      finally
        GraphicStream.Free;
      end
    end
  end
  else
  begin
    WriteString('<EMPTY>');
  end;
end;

procedure TTypedFiler.WriteShortInt(const AValue : ShortInt);
begin
  FStream.WriteBuffer(TS_SHORTINT,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteSingle(const AValue: Single);
begin
  FStream.WriteBuffer(TS_SINGLE,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteSmallInt(const AValue : SmallInt);
begin
  FStream.WriteBuffer(TS_SMALLINT,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteStream(AStream : TStream; FromStart : 
        boolean);
var
  ASize: Int64;
begin
  FStream.WriteBuffer(TS_STREAM,Sizeof(TS_CONST_TYPE));
  if FromStart then
  begin
    AStream.seek(0,soFromBeginning);
  end;
  ASize := AStream.size - AStream.position;
  WriteInt64(ASize);
  
  FStream.CopyFrom(AStream,ASize);
end;

procedure TTypedFiler.WriteString(const AString : string);
var
  ALen: Cardinal;
begin
  FStream.WriteBuffer(TS_STRING,Sizeof(TS_CONST_TYPE));
  ALen := Length(AString);
  WriteCardinal(ALen);
  FStream.Write(AString[1],ALen);
end;

procedure TTypedFiler.WriteVariant(const AValue: Variant);
var
  ItteratorArray: array of integer;
  ArrayHighBounds: array of integer;
  ArrayLowBounds: array of integer;
  ArrayDims: Integer;
  
    Procedure WriteArray(DimNo : integer);
    var
      CellValue : variant;
      i : integer;
    begin
      if (DimNo < ArrayDims) then
      begin
        for i := ArrayLowBounds[DimNo] to ArrayHighBounds[DimNo] do
        begin
          ItteratorArray[DimNo] := i;
          if DimNo = (ArrayDims - 1) then
          begin
            Cellvalue := VarArrayGet(AValue, ItteratorArray);
            WriteVariant(CellValue);
          end
          else
          begin
            WriteArray(DimNo+1);
          end;
        end;
      end;
    end;
  
  var
    VType : TVarType;
    CustomType: TCustomVariantType;
    VarStreamer: IVarStreamable;
    CustomVarStream : TMemoryStream;
    i : integer;
  
begin
  FStream.WriteBuffer(TS_VARIANT,Sizeof(TS_CONST_TYPE));
  VType := VarType(AValue);
  writeVarType(VType); //Store the variant type

  if VarIsArray(AValue,true) then
  begin

    // is an array
    ArrayDims := VarArrayDimCount(AValue);
    WriteInteger(ArrayDims); // save dim count to stream

    // set the low and high array sizes
    Setlength(ArrayLowBounds,ArrayDims);
    Setlength(ArrayHighBounds,ArrayDims);

    // store the low and high bounds
    for i := 0 to ArrayDims-1 do
    begin
      ArrayLowBounds[i] := VarArrayLowBound(AValue,i+1);
      WriteInteger(ArrayLowBounds[i]); // save Low bound to stream
      ArrayHighBounds[i] := VarArrayHighBound(AValue,i+1); // save High bound to stream
      WriteInteger(ArrayHighBounds[i]); // save HighBound To Stream
    end;

    // setup itterator array
    SetLength(ItteratorArray,ArrayDims);
    // initialize to Low Bounds
    for i := 0 to ArrayDims -1 do
    begin
      ItteratorArray[i] := ArrayLowBounds[i];
    end;

    // Write Array Values
    WriteArray(0);

  end
  else
  begin
    // Not an array
    case VType and varTypeMask of // only write basic types
      varEmpty :
      begin
      end;

      varNull:
      begin
      end;

      varOleStr:
      begin
        WriteWideString(AValue);
      end;

      varString:
      begin
        WriteString(AValue);
      end;

      varByte :
      begin
        WriteByte(AValue);
      end;

      varShortInt:
      begin
        WriteShortInt(AValue);
      end;

      varWord:
      begin
        WriteWord(AValue);
      end;

      varSmallInt:
      begin
        WriteSmallInt(AValue);
      end;

      varInteger:
      begin
        WriteInteger(AValue);
      end;

      varSingle:
      begin
        WriteSingle(AValue);
      end;

      varDouble:
      begin
        WriteFloat(AValue)
      end;

      varCurrency:
      begin
        WriteCurrency(AValue)
      end;

      varDate:
      begin
        WriteDateTime(AValue);
      end;

      varBoolean:
      begin
        WriteBoolean(AValue)
      end;

      varLongWord, varInt64:
      begin
        WriteInt64(AValue);
      end;
    else
      begin
        // if it is a custom variant and supports the IVarStreamable interface
        if not FindCustomVariantType(TVarData(AValue).VType, CustomType) or
           not Supports(AValue, IVarStreamable, VarStreamer) then
          WriteString(AValue)
        else
        begin
          CustomVarStream := TMemoryStream.Create;
          try
            // Output custom variant to a stream
            VarStreamer.StreamOut(TVarData(AValue), CustomVarStream);
            // write it's class name
            WriteString(CustomType.ClassName);
            // write the stream
            WriteStream(CustomVarStream,true);
          finally
            CustomVarStream.free;
          end;
        end;
      end;
    end;
  end;
end;

procedure TTypedFiler.WriteVarType(AVarType : TVarType);
begin
  FStream.WriteBuffer(TS_VARTYPE,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AVarType, SizeOf(AVarType));
end;

procedure TTypedFiler.WriteVersion(AValue : Cardinal);
begin
  FStream.WriteBuffer(TS_VERSION,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;

procedure TTypedFiler.WriteWideString(const AWideString : WideString);
var
  ALen: Cardinal;
begin
  FStream.WriteBuffer(TS_WIDESTRING,Sizeof(TS_CONST_TYPE));
  ALen := Length(AWideString)*2;
  WriteCardinal(ALen);
  FStream.Write(Pointer(AWideString)^,ALen);
end;

procedure TTypedFiler.WriteWord(const AValue : Word);
begin
  FStream.WriteBuffer(TS_WORD,Sizeof(TS_CONST_TYPE));
  FStream.WriteBuffer(AValue, SizeOf(AValue));
end;



constructor TTypedFiler.Create(AStream: TStream);
begin
  if not assigned(AStream) then raise exception.create('TTypedFiler cannot be assigned to a null stream');
  FStream := AStream;
end;

destructor TTypedFiler.Destroy;
begin
  inherited;
end;

function TTypedFiler.ReadColor: TColor;
var
  ID: TS_CONST_TYPE;
begin
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_COLOR then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_COLOR)]);

  FStream.ReadBuffer(Result, SizeOf(Result));
end;

{$IFDEF TurboPowerAbbrevia}
procedure TTypedFiler.ReadCompressedStream(Into: TStream);
var
  ID: TS_CONST_TYPE;
begin
  // goto start of stream
  if FStream.Position = FStream.Size then raise exception.Create('End of Stream, cannot read Type');
  FStream.ReadBuffer(ID,Sizeof(TS_CONST_TYPE));
  if ID <> TS_COMPRESSEDSTREAM then raise exception.CreateFmt('Type %s found and %s expected in stream.',[TypedFilerTypeToString(ID),TypedFilerTypeToString(TS_COMPRESSEDSTREAM)]);

  InflateStream(FStream, Into);
end;
{$ELSE}
procedure TTypedFiler.ReadCompressedStream(Into: TStream);
begin
  raise exception.create('TTypedFiler has not been compiled with compression support');
end;
{$ENDIF}

{$IFDEF TurboPowerAbbrevia}
procedure TTypedFiler.WriteCompressedStream(AStream: TStream;
  FromStart: boolean);
begin
  if FromStart then AStream.Seek(0,soFromBeginning);
  FStream.WriteBuffer(TS_COMPRESSEDSTREAM,Sizeof(TS_CONST_TYPE));
  DeflateStream(AStream,FStream);
end;
{$ELSE}
procedure TTypedFiler.WriteCompressedStream(AStream: TStream;
  FromStart: boolean);
begin
  raise exception.create('TTypedFiler has not been compiled with compression support');
end;
{$ENDIF}

function TTypedFiler.RawRead(NumberOfBytes: integer): TDynamicByteArray;
begin
  SetLength(result, NumberOfBytes);
  FStream.Read(pointer(result)^,NumberOfBytes);
end;

procedure TTypedFiler.RawWrite(Data: TDynamicByteArray);
begin
  FStream.Write(pointer(Data)^,High(Data)+1);
end;

end.
